import os
import errno
import os.path as op
from itertools import chain
from .base import Repository
import stat
import shutil
import logging
import json
from fasteners import InterProcessLock
from arcana.data import Fileset, Field
from arcana.pipeline.provenance import Record
from arcana.exceptions import (
    ArcanaError, ArcanaUsageError,
    ArcanaRepositoryError,
    ArcanaMissingDataException,
    ArcanaInsufficientRepoDepthError)
from arcana.utils import get_class_info, HOSTNAME, split_extension


logger = logging.getLogger('arcana')


class BasicRepo(Repository):
    """
    A 'Repository' class for data stored simply in file-system
    directories. Can be a single directory if it contains only one subject
    and visit, otherwise if sub-directories are present (that aren't
    recognised as single filesets) then they are assumed to be
    separate subjects. For multi-visit datasets an additional
    layer of sub-directories for each visit is required within each
    subject sub-directory.

    Parameters
    ----------
    root_dir : str (path)
        Path to local directory containing data
    depth : int
        The number of sub-directory layers under the base directory. For
        example, if depth == 0, then the base directory contains the
        data files and data, if depth == 1, then there is a layer of
        sub-directories for each subject, and if depth == 2 there is
        an additional layer of sub-directories for each visit of each
        subject.
    """

    type = 'directory'
    SUMMARY_NAME = '__ALL__'
    FIELDS_FNAME = 'fields.json'
    PROV_DIR = '__prov__'
    LOCK_SUFFIX = '.lock'
    DEFAULT_SUBJECT_ID = 'SUBJECT'
    DEFAULT_VISIT_ID = 'VISIT'
    MAX_DEPTH = 2

    def __init__(self, root_dir, depth=None, **kwargs):
        super(BasicRepo, self).__init__(**kwargs)
        if not op.exists(root_dir):
            raise ArcanaError(
                "Base directory for BasicRepo '{}' does not exist"
                .format(root_dir))
        self._root_dir = op.abspath(root_dir)
        if depth is None:
            depth = self.guess_depth(root_dir)
        self._depth = depth

    def __repr__(self):
        return "{}(root_dir='{}')".format(type(self).__name__,
                                          self.root_dir)

    def __eq__(self, other):
        try:
            return self.type == other.type and self.root_dir == other.root_dir
        except AttributeError:
            return False

    @property
    def prov(self):
        return {
            'type': get_class_info(type(self)),
            'root': self.root_dir,
            'host': HOSTNAME}

    def __hash__(self):
        return hash(self.type) ^ hash(self.root_dir)

    @property
    def root_dir(self):
        return self._root_dir

    @property
    def depth(self):
        return self._depth

    def get_fileset(self, fileset):
        """
        Set the path of the fileset from the repository
        """
        # Don't need to cache fileset as it is already local as long
        # as the path is set
        if fileset._path is None:
            primary_path = self.fileset_path(fileset)
            aux_files = fileset.format.default_aux_file_paths(primary_path)
            if not op.exists(primary_path):
                raise ArcanaMissingDataException(
                    "{} does not exist in {}"
                    .format(fileset, self))
            for aux_name, aux_path in aux_files.items():
                if not op.exists(aux_path):
                    raise ArcanaMissingDataException(
                        "{} is missing '{}' side car in {}"
                        .format(fileset, aux_name, self))
        else:
            primary_path = fileset.path
            aux_files = fileset.aux_files
        return primary_path, aux_files

    def get_field(self, field):
        """
        Update the value of the field from the repository
        """
        # Load fields JSON, locking to prevent read/write conflicts
        # Would be better if only checked if locked to allow
        # concurrent reads but not possible with multi-process
        # locks (in my understanding at least).
        fpath = self.fields_json_path(field)
        try:
            with InterProcessLock(fpath + self.LOCK_SUFFIX,
                                  logger=logger), open(fpath, 'r') as f:
                dct = json.load(f)
            val = dct[field.name]
            if field.array:
                val = [field.dtype(v) for v in val]
            else:
                val = field.dtype(val)
        except (KeyError, IOError) as e:
            try:
                # Check to see if the IOError wasn't just because of a
                # missing file
                if e.errno != errno.ENOENT:
                    raise
            except AttributeError:
                pass
            raise ArcanaMissingDataException(
                "{} does not exist in the local repository {}"
                .format(field.name, self))
        return val

    def put_fileset(self, fileset):
        """
        Inserts or updates a fileset in the repository
        """
        target_path = self.fileset_path(fileset)
        if op.isfile(fileset.path):
            shutil.copyfile(fileset.path, target_path)
            # Copy side car files into repository
            for aux_name, aux_path in fileset.format.default_aux_file_paths(
                    target_path).items():
                shutil.copyfile(self.aux_file[aux_name], aux_path)
        elif op.isdir(fileset.path):
            if op.exists(target_path):
                shutil.rmtree(target_path)
            shutil.copytree(fileset.path, target_path)
        else:
            assert False

    def put_field(self, field):
        """
        Inserts or updates a field in the repository
        """
        fpath = self.fields_json_path(field)
        # Open fields JSON, locking to prevent other processes
        # reading or writing
        with InterProcessLock(fpath + self.LOCK_SUFFIX, logger=logger):
            try:
                with open(fpath, 'r') as f:
                    dct = json.load(f)
            except IOError as e:
                if e.errno == errno.ENOENT:
                    dct = {}
                else:
                    raise
            if field.array:
                dct[field.name] = list(field.value)
            else:
                dct[field.name] = field.value
            with open(fpath, 'w') as f:
                json.dump(dct, f, indent=2)

    def put_record(self, record):
        fpath = self.prov_json_path(record)
        if not op.exists(op.dirname(fpath)):
            os.mkdir(op.dirname(fpath))
        record.save(fpath)

    def find_data(self, subject_ids=None, visit_ids=None, **kwargs):
        """
        Find all data within a repository, registering filesets, fields and
        provenance with the found_fileset, found_field and found_provenance
        methods, respectively

        Parameters
        ----------
        subject_ids : list(str)
            List of subject IDs with which to filter the tree with. If
            None all are returned
        visit_ids : list(str)
            List of visit IDs with which to filter the tree with. If
            None all are returned

        Returns
        -------
        filesets : list[Fileset]
            All the filesets found in the repository
        fields : list[Field]
            All the fields found in the repository
        records : list[Record]
            The provenance records found in the repository
        """
        all_filesets = []
        all_fields = []
        all_records = []
        for session_path, dirs, files in os.walk(self.root_dir):
            relpath = op.relpath(session_path, self.root_dir)
            path_parts = relpath.split(op.sep) if relpath != '.' else []
            ids = self._extract_ids_from_path(path_parts, dirs, files)
            if ids is None:
                continue
            subj_id, visit_id, from_study = ids
            # Check for summaries and filtered IDs
            if subj_id == self.SUMMARY_NAME:
                subj_id = None
            elif subject_ids is not None and subj_id not in subject_ids:
                continue
            if visit_id == self.SUMMARY_NAME:
                visit_id = None
            elif visit_ids is not None and visit_id not in visit_ids:
                continue
            # Map IDs into ID space of study
            subj_id = self.map_subject_id(subj_id)
            visit_id = self.map_visit_id(visit_id)
            # Determine frequency of session|summary
            if (subj_id, visit_id) == (None, None):
                frequency = 'per_study'
            elif subj_id is None:
                frequency = 'per_visit'
            elif visit_id is None:
                frequency = 'per_subject'
            else:
                frequency = 'per_session'
            filtered_files = self._filter_files(files, session_path)
            for fname in filtered_files:
                basename = split_extension(fname)[0]
                all_filesets.append(
                    Fileset.from_path(
                        op.join(session_path, fname),
                        frequency=frequency,
                        subject_id=subj_id, visit_id=visit_id,
                        repository=self,
                        from_study=from_study,
                        potential_aux_files=[
                            f for f in filtered_files
                            if (split_extension(f)[0] == basename and
                                f != fname)],
                        **kwargs))
            for fname in self._filter_dirs(dirs, session_path):
                all_filesets.append(
                    Fileset.from_path(
                        op.join(session_path, fname),
                        frequency=frequency,
                        subject_id=subj_id, visit_id=visit_id,
                        repository=self,
                        from_study=from_study,
                        **kwargs))
            if self.FIELDS_FNAME in files:
                with open(op.join(session_path,
                                  self.FIELDS_FNAME), 'r') as f:
                    dct = json.load(f)
                all_fields.extend(
                    Field(name=k, value=v, frequency=frequency,
                          subject_id=subj_id, visit_id=visit_id,
                          repository=self, from_study=from_study,
                          **kwargs)
                    for k, v in list(dct.items()))
            if self.PROV_DIR in dirs:
                if from_study is None:
                    raise ArcanaRepositoryError(
                        "Found provenance directory in session directory (i.e."
                        " not in study-specific sub-directory)")
                base_prov_dir = op.join(session_path, self.PROV_DIR)
                for fname in os.listdir(base_prov_dir):
                    all_records.append(Record.load(
                        split_extension(fname)[0],
                        frequency, subj_id, visit_id, from_study,
                        op.join(base_prov_dir, fname)))
        return all_filesets, all_fields, all_records

    def _extract_ids_from_path(self, path_parts, dirs, files):
        depth = len(path_parts)
        if depth == self._depth:
            # Load input data
            from_study = None
        elif (depth == (self._depth + 1) and
              self.PROV_DIR in dirs):
            # Load study output
            from_study = path_parts.pop()
        elif (depth < self._depth and
              any(not f.startswith('.') for f in files)):
            # Check to see if there are files in upper level
            # directories, which shouldn't be there (ignoring
            # "hidden" files that start with '.')
            raise ArcanaRepositoryError(
                "Files ('{}') not permitted at {} level in local "
                "repository".format("', '".join(files),
                                    ('subject'
                                     if depth else 'project')))
        else:
            # Not a directory that contains data files or directories
            return None
        if len(path_parts) == 2:
            subj_id, visit_id = path_parts
        elif len(path_parts) == 1:
            subj_id = path_parts[0]
            visit_id = self.DEFAULT_VISIT_ID
        else:
            subj_id = self.DEFAULT_SUBJECT_ID
            visit_id = self.DEFAULT_VISIT_ID
        return subj_id, visit_id, from_study

    def fileset_path(self, item, fname=None):
        if fname is None:
            fname = item.fname
        subject_id = self.inv_map_subject_id(item.subject_id)
        visit_id = self.inv_map_visit_id(item.visit_id)
        if item.frequency == 'per_study':
            subj_dir = self.SUMMARY_NAME
            visit_dir = self.SUMMARY_NAME
        elif item.frequency.startswith('per_subject'):
            if self.depth < 2:
                raise ArcanaInsufficientRepoDepthError(
                    "Basic repo needs to have depth of 2 (i.e. sub-directories"
                    " for subjects and visits) to hold 'per_subject' data")
            subj_dir = str(subject_id)
            visit_dir = self.SUMMARY_NAME
        elif item.frequency.startswith('per_visit'):
            if self.depth < 1:
                raise ArcanaInsufficientRepoDepthError(
                    "Basic repo needs to have depth of at least 1 (i.e. "
                    "sub-directories for subjects) to hold 'per_visit' data")
            subj_dir = self.SUMMARY_NAME
            visit_dir = str(visit_id)
        elif item.frequency.startswith('per_session'):
            subj_dir = str(subject_id)
            visit_dir = str(visit_id)
        else:
            assert False, "Unrecognised frequency '{}'".format(
                item.frequency)
        if self.depth == 2:
            acq_dir = op.join(self.root_dir, subj_dir, visit_dir)
        elif self.depth == 1:
            acq_dir = op.join(self.root_dir, subj_dir)
        elif self.depth == 0:
            acq_dir = self.root_dir
        else:
            assert False
        if item.from_study is None:
            sess_dir = acq_dir
        else:
            # Append study-name to path (i.e. make a sub-directory to
            # hold derived products)
            sess_dir = op.join(acq_dir, item.from_study)
        # Make session dir if required
        if item.derived and not op.exists(sess_dir):
            os.makedirs(sess_dir, stat.S_IRWXU | stat.S_IRWXG)
        return op.join(sess_dir, fname)

    def fields_json_path(self, field):
        return self.fileset_path(field, self.FIELDS_FNAME)

    def prov_json_path(self, record):
        return self.fileset_path(
            record,
            fname=op.join(self.PROV_DIR, record.pipeline_name + '.json'))

    def guess_depth(self, root_dir):
        """
        Try to guess the depth of a directory repository (i.e. whether it has
        sub-folders for multiple subjects or visits, depending on where files
        and/or derived label files are found in the hierarchy of
        sub-directories under the root dir.

        Parameters
        ----------
        root_dir : str
            Path to the root directory of the repository
        """
        deepest = -1
        for path, dirs, files in os.walk(root_dir):
            depth = self.path_depth(path)
            filtered_files = self._filter_files(files, path)
            if filtered_files:
                logger.info("Guessing depth of directory repository at '{}' is"
                            " {} due to unfiltered files ('{}') in '{}'"
                            .format(root_dir, depth,
                                    "', '".join(filtered_files), path))
                return depth
            if self.PROV_DIR in dirs:
                depth_to_return = max(depth - 1, 0)
                logger.info("Guessing depth of directory repository at '{}' is"
                            "{} due to \"Derived label file\" in '{}'"
                            .format(root_dir, depth_to_return, path))
                return depth_to_return
            if depth >= self.MAX_DEPTH:
                logger.info("Guessing depth of directory repository at '{}' is"
                            " {} as '{}' is already at maximum depth"
                            .format(root_dir, self.MAX_DEPTH, path))
                return self.MAX_DEPTH
            try:
                for fpath in chain(filtered_files,
                                   self._filter_dirs(dirs, path)):
                    Fileset.from_path(fpath)
            except ArcanaError:
                pass
            else:
                if depth > deepest:
                    deepest = depth
        if deepest == -1:
            raise ArcanaRepositoryError(
                "Could not guess depth of '{}' repository as did not find "
                "a valid session directory within sub-directories."
                .format(root_dir))
        return deepest

    @classmethod
    def _filter_files(cls, files, base_dir):
        # Filter out hidden files (i.e. starting with '.')
        return [op.join(base_dir, f) for f in files
                 if not (f.startswith('.') or
                         f.startswith(cls.FIELDS_FNAME))]

    @classmethod
    def _filter_dirs(cls, dirs, base_dir):
        # Filter out hidden directories (i.e. starting with '.')
        # and derived study directories from fileset names
        filtered = [
            op.join(base_dir, d) for d in dirs
            if not (d.startswith('.') or d == cls.PROV_DIR or (
                cls.PROV_DIR in os.listdir(op.join(base_dir, d))))]
        return filtered

    def path_depth(self, dpath):
        relpath = op.relpath(dpath, self.root_dir)
        if '..' in relpath:
            raise ArcanaUsageError(
                "Path '{}' is not a sub-directory of '{}'".format(
                    dpath, self.root_dir))
        elif relpath == '.':
            depth = 0
        else:
            depth = relpath.count(op.sep) + 1
        return depth



from builtins import next
import os.path
from arcana.utils.testing import BaseTestCase
from nipype.pipeline import engine as pe
from arcana.utils.interfaces import ZipDir, UnzipDir


class TestUtilsInterface(BaseTestCase):

    def setUp(self):
        self.reset_dirs()
        self.test_path = os.path.join(self.project_dir, 'test_dir')
        os.mkdir(self.test_path)
        with open(os.path.join(self.test_path, 'test_file'), 'w') as f:
            f.write('test')

    def test_zip_unzip(self):
        # Create Zip node
        zipnode = pe.Node(ZipDir(), name='zip')
        zipnode.inputs.dirname = self.test_path
        # Create Unzip node
        unzipnode = pe.Node(UnzipDir(), name='unzip')
        # Create workflow
        workflow = pe.Workflow('test_zip', base_dir=self.work_dir)
        workflow.connect(zipnode, 'zipped', unzipnode, 'zipped')
        exc_graph = workflow.run()
        unzip_results = next(n for n in exc_graph.nodes()
                             if n.name == 'unzip').result
        self.assertEqual(
            os.listdir(unzip_results.outputs.unzipped), ['test_file'])



from __future__ import division
from builtins import object
import math
import logging
from arcana.exceptions import (
    ArcanaUsageError, ArcanaVersionNotDetectableError, ArcanaVersionError)
import re


logger = logging.getLogger('arcana')


class Version(object):
    """
    Representation of a requirement version. Parses version strings that
    follow the convention

    i.e. <MACRO>.<MINOR>[<PRERELEASE>].[<MICRO>[<PRERELEASE>]][.dev<REVISION>]

    Parameters
    ----------
    requirement : Requirement
        The requirement the version is of
    version : str
        The string representation of the version
    local_name : str
        The name of the requirement as referred to in the local environment
    local_version : str
        The version str as referred to in the local environment
    """

    delimeter = '.'

    def __init__(self, requirement, version, local_name=None,
                 local_version=None):
        self._req = requirement
        self._seq, self._prerelease, self._post, self._dev = self.parse(
            version)
        self._local_version = (
            local_version if local_version != str(self) else None)
        self._local_name = (
            local_name if local_name != requirement.name else None)

    @property
    def requirement(self):
        return self._req

    @property
    def name(self):
        return self.requirement.name

    @property
    def sequence(self):
        return self._seq

    @property
    def prerelease(self):
        return self._prerelease

    @property
    def dev(self):
        return self._dev

    @property
    def post(self):
        return self._post

    @property
    def local_version(self):
        """
        The un-parsed version string. Used to cross-reference with list of
        available versions coming from an Environment
        """
        return (self._local_version
                if self._local_version is not None else str(self))

    @property
    def local_name(self):
        return (self._local_name
                if self._local_name is not None else self.requirement.name)

    def serialise(self):
        pass

    @classmethod
    def unserialise(cls, dct):
        pass

    def __str__(self):
        s = self.delimeter.join(str(i) for i in self._seq)
        if self._prerelease is not None:
            s += '{}{}'.format(*self._prerelease)
        if self._post is not None:
            s += '.post{}'.format(self._post)
        if self._dev is not None:
            s += '.dev{}'.format(self._dev)
        return s

    def __repr__(self):
        return "{}[{}]".format(self._req, str(self))

    def __eq__(self, other):
        return (self._req == other._req and
                self._seq == other._seq and
                self._prerelease == other._prerelease and
                self._dev == other._dev)

    def compare(self, other):
        """
        Compares the version with another

        Parameters
        ----------
        other : Version
            The version to compare to
        """
        if self._req != other._req:
            raise ArcanaUsageError(
                "Can't compare versions of different requirements {} and {}"
                .format(self._req, other._req))
        # Compare main sequence
        if self._seq < other._seq:
            return -1
        elif self._seq > other._seq:
            return 1
        # If main sequence is equal check prerelease. If a prerelease is
        # None then it is a full release which is > then a prerelease so we
        # just assign it 'z' (which is greater than 'a', 'b' and 'rc')
        s = self._prerelease if self._prerelease is not None else ('z',)
        o = other._prerelease if other._prerelease is not None else ('z',)
        if s < o:
            return -1
        if s > o:
            return 1
        # If both main sequence and prereleases are equal, compare post release
        s = self._post if self._post is not None else 0
        o = other._post if other._post is not None else 0
        if s < o:
            return -1
        if s > o:
            return 1
        # If both main sequence and prereleases are equal, compare development
        # release
        s = self._dev if self._dev is not None else 0
        o = other._dev if other._dev is not None else 0
        if s < o:
            return -1
        if s > o:
            return 1
        assert self == other
        return 0

    def __lt__(self, other):
        return self.compare(other) < 0

    def __gt__(self, other):
        return self.compare(other) > 0

    def __le__(self, other):
        return self.compare(other) <= 0

    def __ge__(self, other):
        return self.compare(other) >= 0

    @property
    def regex(self):
        # Escape delimeter if required
        m = ('\\' + self.delimeter if self.delimeter in r'{}\$.|?*+()[]'
             else self.delimeter)
        # Pattern to match sub-version
        sub_ver = r'{}\d+[a-zA-Z\-_0-9]*'.format(m)
        return re.compile(
            r'(?<!\d{m})(\d+{sv}(?:{sv})?(?:{m}\w+)?)'.format(m=m, sv=sub_ver))

    def parse(self, version):
        """
        Splits a typical version string (e.g. <MAJOR>.<MINOR>.<MICRO>)
        into a tuple that can be sorted properly. Ignores all leading
        and trailing characters by using a regex search (instead of match) so
        as to pick the version string out of a block of text.

        Parameters
        ----------
        version : str | int | float | tuple(int)
            The string containing the version numbers, or alternatively an
            integer, float (number after decimal is interpreted as minor ver),
            or tuple|list containing the version sequence.

        Returns
        -------
        sequence : tuple(int | str)
            A tuple containing the main sequence of the version,
            e.g. <MAJOR>.<MINOR>.<MICRO>
        prerelease : 2-tuple(str, int) | None
            A 2-tuple containing the type of prerelease ('a' - alpha,
            'b' - beta, or 'rc' - release-canditate) and the number of the
            prerelease
        post : int |None
            The number of the post version
        dev : int | None
            The number of the development version
        """
        # Check to see if version is not a string but rather another type
        # that can be interpreted as a version
        if isinstance(version, int):
            return (version,), None, None, None
        elif isinstance(version, (tuple, list)):
            return tuple(int(i) for i in version), None, None, None
        elif isinstance(version, float):
            major = math.floor(version)
            minor = version - major
            return (major, minor), None, None, None
        match = self.regex.search(version)
        if match is None:
            raise ArcanaVersionNotDetectableError(
                "Could not parse version string {} as {}. Regex ({}) did not "
                "match any sub-string".format(version, type(self).__name__,
                                              self.regex.pattern))
        sequence = []
        prerelease = None
        dev = None
        post = None
        for part in match.group(1).split(self.delimeter):
            if part.startswith('dev'):
                dev = int(part[len('dev'):])
            elif part.startswith('post'):
                post = int(part[len('post'):])
            else:
                # Split on non-numeric parts of the version string so that we
                # can detect prerelease
                sub_parts = re.split('([^\d]+)', part)
                if sub_parts[0]:
                    try:
                        seq_part = int(sub_parts[0])
                    except ValueError:
                        seq_part = sub_parts[0]
                    sequence.append(seq_part)
                if len(sub_parts) > 1:
                    stage = sub_parts[1]
                    try:
                        pr_ver = int(sub_parts[2])
                    except ValueError:
                        raise ArcanaVersionNotDetectableError(
                            "Could not parse version string {} as {}. "
                            "Did not recognise pre-release version {}"
                            .format(version, type(self).__name__,
                                    sub_parts[2]))
                    stage = stage.strip('-_').lower()
                    if not stage:  # No prerelease info, assume a dev version
                        assert dev is None
                        dev = pr_ver
                        continue
                    if 'alpha'.startswith(stage):
                        stage = 'a'
                    elif 'beta'.startswith(stage):
                        stage = 'b'
                    elif stage == 'rc' or stage == 'release-canditate':
                        stage = 'rc'
                    else:
                        raise ArcanaVersionNotDetectableError(
                            "Could not parse version string {} as {}. "
                            "Did not recognise pre-release stage {}"
                            .format(version, type(self).__name__, stage))
                    prerelease = (stage, pr_ver)
        return tuple(sequence), prerelease, post, dev

    def within(self, version):
        """
        A single version can also be interpreted as an open range (i.e. no
        maximum version)
        """
        if not isinstance(version, Version):
            version = type(self._min_ver)(self._req, version)
        return version >= self

    def latest_within(self, *args, **kwargs):
        return self._req.latest_within_range(self, *args, **kwargs)

    @property
    def prov(self):
        prov = {'version': str(self)}
        if self._local_name is not None:
            prov['local_name'] = self.local_name
        if self._local_version is not None:
            prov['local_version'] = self.local_version
        return prov


class VersionRange(object):
    """
    A range of versions associated with a software requirement

    Parameters
    ----------
    min_version : Version
        The minimum version required by the node
    max_version : Version
        The maximum version that is compatible with the Node
    """

    def __init__(self, min_version, max_version):
        if min_version.requirement != max_version.requirement:
            raise ArcanaUsageError(
                "Inconsistent requirements between min and max versions "
                "({} and {})".format(min_version.requirement,
                                     max_version.requirement))
        self._min_ver = min_version
        self._max_ver = max_version
        if max_version < min_version:
            raise ArcanaUsageError(
                "Maxium version in is less than minimum in {}"
                .format(self))

    @property
    def name(self):
        return self.minimum.name

    @property
    def requirement(self):
        return self.minimum.requirement

    @property
    def minimum(self):
        return self._min_ver

    @property
    def maximum(self):
        return self._max_ver

    def __eq__(self, other):
        return (self._min_ver == other._min_ver and
                self._max_ver == other._max_ver)

    def __str__(self):
        return '{} <= v <= {}'.format(self.minimum, self.maximum)

    def __repr__(self):
        return "{}[{}]".format(
            self._min_ver.requirement, self)

    def within(self, version):
        if not isinstance(version, Version):
            version = type(self._min_ver)(self.requirement, version)
        return version >= self._min_ver and version <= self._max_ver

    def latest_within(self, *args, **kwargs):
        return self._min_ver.requirement.latest_within_range(self, *args,
                                                             **kwargs)


class BaseRequirement(object):
    """
    Base class for a details of a software package that is required by
    a node of a pipeline.

    Parameters
    ----------
    name : str
        Name of the package
    references : list[Citation]
        A list of references that should be cited when using this software
        requirement
    website : str
        Address of the website detailing the software
    delimeter : str
        Delimeter used to split a version string
    """

    def __init__(self, name, citations=None, website=None,
                 version_cls=Version):
        self._name = name.lower()
        self._citations = citations if citations is not None else []
        self._website = website
        self._version_cls = version_cls

    def __eq__(self, other):
        return (self.name == other.name and self.website == other.website)

    @property
    def name(self):
        return self._name

    @property
    def version_cls(self):
        return self._version_cls

    def __hash__(self):
        return hash(self.name) ^ hash(self.website)

    def __repr__(self):
        return "{}(name={})".format(type(self).__name__, self.name)

    def v(self, version, max_version=None, **kwargs):
        """
        Returns either a single requirement version or a requirement version
        range depending on whether two arguments are supplied or one

        Parameters
        ----------
        version : str | Version
            Either a version of the requirement, or the first version in a
            range of acceptable versions
        """
        if not isinstance(version, Version):
            version = self.version_cls(self, version, **kwargs)
        # Return a version range instead of version
        if max_version is not None:
            if not isinstance(max_version, Version):
                max_version = self.version_cls(self, max_version, **kwargs)
            version = VersionRange(version, max_version)
        return version

    @property
    def citations(self):
        return iter(self._citations)

    @property
    def website(self):
        return self._website

    def detect_version(self, **kwargs):
        return self.version_cls(self, self.detect_version_str(), **kwargs)

    def detect_version_str(self):
        """
        Detects and returns the version string of the software requirement
        that is accessible in the current environment. NB: to be overridden in
        sub-classes.

        * If the requirement is not available in the current environment:
            raise ArcanaRequirementNotFoundError
        * If the requirement is available but its version cannot be detected
          for whatever reason:
            raise ArcanaVersionNotDetectableError
        """
        raise NotImplementedError

    def latest_within_range(self, version_range, available):
        """
        Picks the latest acceptible version from the versions available

        Parameters
        ----------
        version_range : VersionRange | Version
            A range of versions or a single version. A single version
            will be interpreted that there are no upper bounds on the version
            range
        available : list(Version)
            List of possible versions to select from

        Returns
        -------
        latest : Version
            The latest version
        """
        latest_ver = None
        for ver in available:
            if version_range.within(ver) and (latest_ver is None or
                                              ver > latest_ver):
                latest_ver = ver
        if latest_ver is None:
            if isinstance(version_range, VersionRange):
                msg_part = 'within range'
            else:
                msg_part = 'greater than'
            raise ArcanaVersionError(
                "Could not find version {} {} from available: {}"
                .format(msg_part, version_range,
                        ', '.join(str(v) for v in available)))
        return latest_ver



from past.builtins import basestring
from builtins import object
from operator import attrgetter
from copy import copy, deepcopy
from arcana.exceptions import (
    ArcanaError, ArcanaUsageError,
    ArcanaOutputNotProducedException,
    ArcanaMissingDataException, ArcanaNameError)
from .base import BaseFileset, BaseField, BaseData
from .item import Fileset, Field
from .collection import FilesetCollection, FieldCollection


class BaseInputSpec(object):

    derived = False
    # For duck-typing with *Input objects
    skip_missing = False
    drop_if_missing = False
    derivable = False

    def __init__(self, name, desc=None, optional=False, default=None):
        if optional and default is not None:
            raise ArcanaUsageError(
                "'optional' doesn't make sense for specs ('{}') with default "
                "values"
                .format(name))
        self._desc = desc
        self._study = None
        self._optional = optional
        # Set the name of the default collection-like object so it matches
        # the name of the spec
        if default is not None:
            if default.frequency != self.frequency:
                raise ArcanaUsageError(
                    "Frequency of default collection-like object passed to "
                    "'{}' spec ('{}'), does not match spec ('{}')".format(
                        name, default.frequency, self.frequency))
            default = deepcopy(default)
        self._default = default

    def __eq__(self, other):
        return (self.desc == other.desc and
                self.optional == other.optional and
                self.default == other.default)

    def __hash__(self):
        return (hash(self.desc) ^ hash(self.optional) ^ hash(self.default))

    def initkwargs(self):
        dct = {}
        dct['desc'] = self.desc
        dct['optional'] = self.optional
        dct['default'] = deepcopy(self.default)
        return dct

    def find_mismatch(self, other, indent=''):
        mismatch = ''
        sub_indent = indent + '  '
        if self.optional != other.optional:
            mismatch += ('\n{}optional: self={} v other={}'
                         .format(sub_indent, self.optional,
                                 other.optional))
        if self.default != other.default:
            mismatch += ('\n{}default: self={} v other={}'
                         .format(sub_indent, self.default,
                                 other.default))
        if self.desc != other.desc:
            mismatch += ('\n{}desc: self={} v other={}'
                         .format(sub_indent, self.desc, other.desc))
        return mismatch

    def bind(self, study, **kwargs):  # @UnusedVariable
        """
        Returns a copy of the InputSpec bound to the given study

        Parameters
        ----------
        study : Study
            A study to bind the fileset spec to (should happen in the
            study __init__)
        """
        if self.default is None:
            raise ArcanaError(
                "Attempted to bind '{}' to {} but only acquired specs with "
                "a default value should be bound to studies{})".format(
                    self.name, study))
        if self._study is not None:
            # This avoids rebinding specs to sub-studies that have already
            # been bound to the multi-study
            bound = self
        else:
            bound = copy(self)
            bound._study = study
            bound._default = bound.default.bind(study)
        return bound

    @property
    def optional(self):
        return self._optional

    @property
    def study(self):
        if self._study is None:
            raise ArcanaError(
                "{} is not bound to a study".format(self))
        return self._study

    @property
    def default(self):
        # Ensure the default has the same name as the spec to it can be
        # accessed properly by the source node
        if self._default is not None:
            # FIXME: Should use setter but throwing an error for some reason
            self._default.name = self.name
        return self._default

    @property
    def desc(self):
        return self._desc

    @property
    def collection(self):
        if self._study is None:
            raise ArcanaUsageError(
                "{} needs to be bound to a study before accessing "
                "the corresponding collection".format(self))
        if self.default is None:
            raise ArcanaUsageError(
                "{} does not have default so cannot access its collection"
                .format(self))
        return self.default.collection


class BaseSpec(object):

    derived = True
    # For duck-typing with *Input objects
    skip_missing = False
    drop_if_missing = False
    derivable = True

    def __init__(self, name, pipeline_getter, desc=None,
                 pipeline_args=None, group=None):
        if pipeline_getter is not None:
            if not isinstance(pipeline_getter, basestring):
                raise ArcanaUsageError(
                    "Pipeline name for {} '{}' is not a string "
                    "'{}'".format(name, pipeline_getter))
        self._pipeline_getter = pipeline_getter
        self._desc = desc
        self._study = None
        self._collection = None
        if isinstance(pipeline_args, dict):
            pipeline_args = tuple(sorted(pipeline_args.items()))
        elif pipeline_args is not None:
            pipeline_args = tuple(pipeline_args)
        else:
            pipeline_args = ()
        self._pipeline_args = pipeline_args
        self._group = group

    def __eq__(self, other):
        return (self.pipeline_getter == other.pipeline_getter and
                self.desc == other.desc and
                self.pipeline_args == other.pipeline_args and
                self.group == other.group)

    def __hash__(self):
        return (hash(self.pipeline_getter) ^ hash(self.desc) ^
                hash(self.pipeline_args.items()) ^ hash(self.group))

    def initkwargs(self):
        dct = {}
        dct['pipeline_getter'] = self.pipeline_getter
        dct['desc'] = self.desc
        dct['pipeline_args'] = copy(self.pipeline_args)
        dct['group'] = self.group
        return dct

    def find_mismatch(self, other, indent=''):
        mismatch = ''
        sub_indent = indent + '  '
        if self.pipeline_getter != other.pipeline_getter:
            mismatch += ('\n{}pipeline_getter: self={} v other={}'
                         .format(sub_indent, self.pipeline_getter,
                                 other.pipeline_getter))
        if self.desc != other.desc:
            mismatch += ('\n{}desc: self={} v other={}'
                         .format(sub_indent, self.desc, other.desc))
        if self.group != other.group:
            mismatch += ('\n{}group: self={} v other={}'
                         .format(sub_indent, self.group, other.group))
        if self.pipeline_args != other.pipeline_args:
            mismatch += ('\n{}pipeline_args: self={} v other={}'
                         .format(sub_indent, self.pipeline_args,
                                 other.pipeline_args))
        return mismatch

    def bind(self, study, **kwargs):  # @UnusedVariable
        """
        Returns a copy of the Spec bound to the given study

        Parameters
        ----------
        study : Study
            A study to bind the fileset spec to (should happen in the
            study __init__)
        """
        if self._study is not None:
            # Avoid rebinding specs in sub-studies that have already
            # been bound to MultiStudy
            bound = self
        else:
            bound = copy(self)
            bound._study = study
            if not hasattr(study, self.pipeline_getter):
                raise ArcanaError(
                    "{} does not have a method named '{}' required to "
                    "derive {}".format(study, self.pipeline_getter,
                                       self))
            bound._bind_tree(study.tree)
        return bound

    @property
    def collection(self):
        if self._collection is None:
            raise ArcanaUsageError(
                "{} needs to be bound to a study before accessing "
                "the corresponding collection".format(self))
        return self._collection

    def nodes(self, tree):
        """
        Returns the relevant nodes for the spec's frequency
        """
        # Run the match against the tree
        if self.frequency == 'per_session':
            nodes = []
            for subject in tree.subjects:
                for sess in subject.sessions:
                    nodes.append(sess)
        elif self.frequency == 'per_subject':
            nodes = tree.subjects
        elif self.frequency == 'per_visit':
            nodes = tree.visits
        elif self.frequency == 'per_study':
            nodes = [tree]
        else:
            assert False, "Unrecognised frequency '{}'".format(
                self.frequency)
        return nodes

    @property
    def derivable(self):
        """
        Whether the spec (only valid for derived specs) can be derived
        given the inputs and switches provided to the study
        """
        try:
            # Just need to iterate all study inputs and catch relevant
            # exceptions
            list(self.pipeline.study_inputs)
        except (ArcanaOutputNotProducedException,
                ArcanaMissingDataException):
            return False
        return True

    @property
    def pipeline_getter(self):
        return self._pipeline_getter

    @property
    def group(self):
        return self._group

    @property
    def pipeline_args(self):
        return self._pipeline_args

    @property
    def pipeline_arg_names(self):
        return tuple(n for n, _ in self._pipeline_args)

    @property
    def pipeline(self):
        return self.study.pipeline(self.pipeline_getter, [self.name],
                                   pipeline_args=self.pipeline_args)

    @property
    def study(self):
        if self._study is None:
            raise ArcanaError(
                "{} is not bound to a study".format(self))
        return self._study

    @property
    def desc(self):
        return self._desc

    def _tree_node(self, subject_id=None, visit_id=None):
        if self.frequency == 'per_session':
            node = self.study.tree.subject(subject_id).session(visit_id)
        elif self.frequency == 'per_subject':
            node = self.study.tree.subject(subject_id)
        elif self.frequency == 'per_visit':
            node = self.study.tree.visit(visit_id)
        elif self.frequency == 'per_study':
            node = self.study.tree
        else:
            assert False
        return node


class InputFilesetSpec(BaseFileset, BaseInputSpec):
    """
    A specification for an "acquired" fileset (e.g from the scanner or
    standard atlas)

    Parameters
    ----------
    name : str
        The name of the fileset
    valid_formats : FileFormat | list[FileFormat]
        The acceptable file formats for input filesets to match this spec
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    desc : str
        Description of what the field represents
    optional : bool
        Whether the specification is optional or not. Only valid for
        "acquired" fileset specs.
    default : FilesetCollection | object
        The default value to be passed as an input to this spec if none are
        provided. Can either be an explicit FilesetCollection or any object
        with a 'collection' property that will return a default collection.
        This object should also implement a 'bind(self, study)' method to
        allow the study to be bound to it.
    pipeline_args : dct[str, *] | None
        Arguments to pass to the pipeline constructor method. Avoids having to
        create separate methods for each spec, where the only difference
        between the specs
    """

    is_spec = True
    CollectionClass = FilesetCollection

    def __init__(self, name, valid_formats, frequency='per_session', # @ReservedAssignment @IgnorePep8
                 desc=None, optional=False, default=None):
        # Ensure allowed formats is a list
        try:
            valid_formats = sorted(valid_formats, key=attrgetter('name'))
        except TypeError:
            valid_formats = [valid_formats]
        else:
            if not valid_formats:
                raise ArcanaError(
                    "'{}' spec doesn't have any allowed formats".format(name))
        self._valid_formats = tuple(valid_formats)
        BaseFileset.__init__(self, name, None, frequency)
        BaseInputSpec.__init__(self, name, desc, optional=optional,
                               default=default)

    @property
    def valid_formats(self):
        return self._valid_formats

    @property
    def format(self):
        try:
            return self.default.format
        except AttributeError:
            raise ArcanaUsageError(
                "File format is not defined for InputFilesetSpec objects "
                "without a default")

    def __eq__(self, other):
        return (BaseFileset.__eq__(self, other) and
                BaseInputSpec.__eq__(self, other) and
                self._valid_formats == other._valid_formats)

    def __hash__(self):
        return (BaseFileset.__hash__(self) ^
                BaseInputSpec.__hash__(self) and
                hash(self.valid_formats))

    def initkwargs(self):
        dct = BaseData.initkwargs(self)
        dct.update(BaseInputSpec.initkwargs(self))
        dct['valid_formats'] = self._valid_formats
        return dct

    def __repr__(self):
        return ("{}(name='{}', valid_formats={}, frequency={}, default={}, "
                "optional={})"
                .format(type(self).__name__, self.name,
                        list(f.name for f in self.valid_formats),
                        self.frequency, self.default, self.optional))

    def find_mismatch(self, other, indent=''):
        sub_indent = indent + '  '
        mismatch = BaseFileset.find_mismatch(self, other, indent)
        mismatch += BaseInputSpec.find_mismatch(self, other, indent)
        if self.valid_formats != other.valid_formats:
            mismatch += ('\n{}pipeline: self={} v other={}'
                         .format(sub_indent, list(self.valid_formats),
                                 list(other.valid_formats)))
        return mismatch


class FilesetSpec(BaseFileset, BaseSpec):
    """
    A specification for a fileset within a study to be derived from a
    processing pipeline.

    Parameters
    ----------
    name : str
        The name of the fileset
    format : FileFormat
        The file format used to store the fileset. Can be one of the
        recognised formats
    pipeline_getter : str
        Name of the method in the study that constructs a pipeline to derive
        the fileset
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    desc : str
        Description of what the field represents
    valid_formats : list[FileFormat]
        A list of valid file formats that can be supplied to the spec if
        overridden as an input. Typically not required, but useful for some
        specs that are typically provided as inputs (e.g. magnitude MRI)
        but can be derived from other inputs (e.g. coil-wise MRI images)
    pipeline_args : dct[str, *] | None
        Arguments to pass to the pipeline constructor method. Avoids having to
        create separate methods for each spec, where the only difference
        between the specs are interface parameterisations
    group : str | None
        A name for a group of fileset specs. Used improve human searching of
        available options
    """

    is_spec = True
    CollectionClass = FilesetCollection

    def __init__(self, name, format, pipeline_getter, frequency='per_session',  # @ReservedAssignment @IgnorePep8 
                 desc=None, valid_formats=None, pipeline_args=None,
                 group=None):
        BaseFileset.__init__(self, name, format, frequency)
        BaseSpec.__init__(self, name, pipeline_getter, desc,
                          pipeline_args, group)
        if valid_formats is not None:
            # Ensure allowed formats is a list
            try:
                valid_formats = sorted(valid_formats, key=attrgetter('name'))
            except TypeError:
                valid_formats = [valid_formats]
        self._valid_formats = valid_formats

    def __eq__(self, other):
        return (BaseFileset.__eq__(self, other) and
                BaseSpec.__eq__(self, other) and
                self.valid_formats == other.valid_formats)

    def __hash__(self):
        return (BaseFileset.__hash__(self) ^
                BaseSpec.__hash__(self) ^
                hash(self.valid_formats))

    def initkwargs(self):
        dct = BaseFileset.initkwargs(self)
        dct.update(BaseSpec.initkwargs(self))
        dct['valid_formats'] = self._valid_formats
        return dct

    def __repr__(self):
        return ("FilesetSpec(name='{}', format={}, pipeline_getter={}, "
                "frequency={})".format(
                    self.name, self.format, self.pipeline_getter,
                    self.frequency))

    def find_mismatch(self, other, indent=''):
        sub_indent = indent + '  '
        mismatch = BaseFileset.find_mismatch(self, other, indent)
        mismatch += BaseSpec.find_mismatch(self, other, indent)
        if self.valid_formats != other.valid_formats:
            mismatch += ('\n{}pipeline: self={} v other={}'
                         .format(sub_indent, list(self.valid_formats),
                                 list(other.valid_formats)))
        return mismatch

    def _bind_node(self, node, **kwargs):
        try:
            fileset = node.fileset(self.name, from_study=self.study.name,
                                   format=self.format)
        except ArcanaNameError:
            # For filesets that can be generated by the analysis
            fileset = Fileset(self.name, format=self.format,
                              frequency=self.frequency, path=None,
                              subject_id=node.subject_id,
                              visit_id=node.visit_id,
                              repository=self.study.repository,
                              from_study=self.study.name,
                              exists=False,
                              **kwargs)
        return fileset

    def _bind_tree(self, tree, **kwargs):
        self._collection = FilesetCollection(
            self.name,
            (self._bind_node(n, **kwargs) for n in self.nodes(tree)),
            frequency=self.frequency,
            format=self.format)

    @property
    def valid_formats(self):
        if self._valid_formats is not None:
            valid_formats = self._valid_formats
        else:
            valid_formats = [self.format] + list(self.format.convertable_from)
        return valid_formats


class InputFieldSpec(BaseField, BaseInputSpec):
    """
    An abstract base class representing an acquired field

    Parameters
    ----------
    name : str
        The name of the fileset
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    desc : str
        Description of what the field represents
    optional : bool
        Whether the specification is optional or not. Only valid for
        "acquired" fileset specs.
    default : FieldCollection | callable
        The default value to be passed as an input to this spec if none are
        provided. Can either be an explicit FieldCollection or any object
        with a 'collection' property that will return a default collection.
        This object should also implement a 'bind(self, study)' method to
        allow the study to be bound to it.
    """

    is_spec = True
    CollectionClass = FieldCollection

    def __init__(self, name, dtype, frequency='per_session', desc=None,
                 optional=False, default=None, array=False):
        BaseField.__init__(self, name, dtype, frequency, array=array)
        BaseInputSpec.__init__(self, name, desc, optional=optional,
                                  default=default)

    def __eq__(self, other):
        return (BaseField.__eq__(self, other) and
                BaseInputSpec.__eq__(self, other))

    def __hash__(self):
        return (BaseField.__hash__(self) ^ BaseInputSpec.__hash__(self))

    def find_mismatch(self, other, indent=''):
        mismatch = BaseField.find_mismatch(self, other, indent)
        mismatch += BaseInputSpec.find_mismatch(self, other, indent)
        return mismatch

    def __repr__(self):
        return ("{}(name='{}', dtype={}, frequency={}, array={})".format(
            self.__class__.__name__, self.name, self.dtype,
            self.frequency, self.array))

    def initkwargs(self):
        dct = BaseField.initkwargs(self)
        dct.update(BaseInputSpec.initkwargs(self))
        return dct


class FieldSpec(BaseField, BaseSpec):
    """
    An abstract base class representing the specification for a derived
    fileset.

    Parameters
    ----------
    name : str
        The name of the fileset
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    pipeline_getter : str
        Name of the method that constructs pipelines to derive the field
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    desc : str
        Description of what the field represents
    pipeline_args : dct[str, *] | None
        Arguments to pass to the pipeline constructor method. Avoids having to
        create separate methods for each spec, where the only difference
        between the specs are interface parameterisations
    group : str
        A name for a group of fileset specs. Used improve human searching of
        available options
    """

    is_spec = True
    CollectionClass = FieldCollection

    def __init__(self, name, dtype, pipeline_getter=None,
                 frequency='per_session', desc=None, array=False,
                 pipeline_args=None, group=None):
        BaseField.__init__(self, name, dtype, frequency, array=array)
        BaseSpec.__init__(self, name, pipeline_getter, desc,
                          pipeline_args=pipeline_args, group=group)

    def __eq__(self, other):
        return (BaseField.__eq__(self, other) and
                BaseSpec.__eq__(self, other))

    def __hash__(self):
        return (BaseField.__hash__(self) ^ BaseSpec.__hash__(self))

    def find_mismatch(self, other, indent=''):
        mismatch = BaseField.find_mismatch(self, other, indent)
        mismatch += BaseSpec.find_mismatch(self, other, indent)
        return mismatch

    def __repr__(self):
        return ("{}(name='{}', dtype={}, pipeline_getter={}, "
                "frequency={}, array={})".format(
                    self.__class__.__name__, self.name, self.dtype,
                    self.pipeline_getter, self.frequency, self.array))

    def initkwargs(self):
        dct = BaseField.initkwargs(self)
        dct.update(BaseSpec.initkwargs(self))
        return dct

    def _bind_node(self, node, **kwargs):
        try:
            field = node.field(self.name, from_study=self.study.name)
        except ArcanaNameError:
            # For fields to be generated by the analysis
            field = Field(self.name, dtype=self.dtype,
                          frequency=self.frequency,
                          subject_id=node.subject_id,
                          visit_id=node.visit_id,
                          repository=self.study.repository,
                          from_study=self.study.name,
                          array=self.array,
                          exists=False, **kwargs)
        return field

    def _bind_tree(self, tree, **kwargs):
        self._collection = FieldCollection(
            self.name,
            (self._bind_node(n, **kwargs) for n in self.nodes(tree)),
            frequency=self.frequency,
            dtype=self.dtype,
            array=self.array)



from .item import Fileset, Field
from .collection import FilesetCollection, FieldCollection
from .spec import (
    FilesetSpec, FieldSpec, InputFilesetSpec, InputFieldSpec, BaseSpec,
    BaseInputSpec)
from .base import BaseField, BaseFileset, BaseData
from .input import InputFilesets, InputFields, BaseInput
from .file_format import FileFormat, Converter, IdentityConverter
